#include "analyze.h"

using namespace std;

// ******************************************************
// Functions to manipulate Analyze 7.5 images 
// MRFSEG + GAMIXTURE software bundle: Voxel classification in volumetric 
// images using genetic algorithms and Markov random fields.
// The genetic algorithm is described in: 
// [1]  J. Tohka , E. Krestyannikov, I.D. Dinov , A. MacKenzie-Graham,
// D.W. Shattuck , U. Ruotsalainen, and A.W. Toga. 
// Genetic algorithms for finite mixture model based voxel classification
// in neuroimaging.  
// IEEE Transactions on Medical Imaging  , 26(5):696 - 711, 2007.
// (C) Jussi Tohka, 2005 - 2007 
// Laboratory of NeuroImaging, Department of Neurology, UCLA Medical School, CA, USA
// Institute of Signal Processing, Tampere University of Technology, Finland  
// Version 1.1
// e-mail jussi.tohka@tut.fi
// *****************************************************
// ******************************************************
// Permission to use, copy, modify, and distribute this software 
// for any purpose and without fee is hereby
// granted, provided that the above copyright notice appears in all
// copies.  The author, Tampere University of Technology and University of
// California, Los Angeles make no representations
// about the suitability of this software for any purpose.  It is
// provided "as is" without express or implied warranty.
// *****************************************************
// (C) Jussi Tohka, 2005 
// Laboratory of NeuroImaging, Department of Neurology, UCLA Medical School, CA, USA
// Institute of Signal Processing, Tampere University of Technology, Finland  
// Version 1.0, for internal distribution only.
// e-mail jussi.tohka@{tut.fi,loni.ucla.edu}
// *****************************************************



// Reads an Analyze image 
// Supports datatypes UNSIGNED_CHAR, SIGNED SHORT, SIGNED INT AND SIGNED FLOAT
// Returns:  0 if succesful
//           1 if the header could not be opened
//           2 if the memory allocation did not work as usual
//           3 if the image file could not be opened
//           4 if the datatype was not supported
//


int readImage(char* filename,AnalyzeImage* img, bool signedData)
{
 
  char imgfname[256];
  unsigned char* tmpchar;
  short* tmpshort;
  unsigned short* tmpushort;
  int* tmpint;  
  int i,imgsize;
  bool bs;
 
 
  if(readHeader(filename,img->header) == false) return(1);
  bs = byteSwapNecessary(&img->header);
 
  if(bs) swapHeader(img->header);  
 
  imgsize = (img->header.x_dim)*(img->header.y_dim)*(img->header.z_dim); 
  // img->data = (float *) malloc(imgsize*sizeof(float));  
  img->data = new float[imgsize];
  if(img->data == NULL) return(2);
  imagename(imgfname,filename);
  ifstream ifile(imgfname,ios::binary);
  if (!ifile) return(3);
  
  switch(img->header.datatype) {
	case DT_UNSIGNED_CHAR:
	     tmpchar = new unsigned char [imgsize];
             ifile.read((char*) tmpchar,imgsize);
             for(i = 0;i < imgsize;i++) {
               img->data[i] =  (float) tmpchar[i];
             }
             delete[] tmpchar;
             ifile.close();
             break;
        case DT_SIGNED_SHORT: 
	     if(signedData) {
	       tmpshort = new short [imgsize];  
               ifile.read((char*)tmpshort,imgsize*sizeof(tmpshort));
               for(i = 0;i < imgsize;i++) {
                 if (bs) byteswap(tmpshort[i]);
                 img->data[i] = (float) tmpshort[i];
               }            
               delete[] tmpshort;
               ifile.close();
             }
             else {
               tmpushort = new unsigned short [imgsize];  
               ifile.read((char*)tmpushort,imgsize*sizeof(tmpshort));
               for(i = 0;i < imgsize;i++) {
                 if (bs) byteswap(tmpushort[i]);
                 img->data[i] = (float) tmpushort[i];
               }            
               delete[] tmpushort;
               ifile.close();
             }
             break;
        case DT_SIGNED_INT: 
	     tmpint = new int [imgsize];  
	     ifile.read((char*) tmpint,imgsize*sizeof(int));
             for(i = 0;i < imgsize;i++) {
               if (bs) swapx(tmpint[i]);
               img->data[i] = (float) tmpint[i];
             }
             ifile.close();
             delete[] tmpint;
             break;  
        case DT_FLOAT:             
	     ifile.read((char*)img->data,imgsize*sizeof(float));
             for(i = 0;i < imgsize;i++) {
               if (bs) swapf((unsigned int *)&img->data[i]);
               
             }
             ifile.close();
             break;
        default: return(4); break;
  }
 
  return(0);

}
   

// Writes an Analyze Image 
// returns 0 if ok
// returns 1 if the header could not be opened for writing
// returns 2 if the image file could not be opened for writing 
// returns 3 if the datatype was not supported
// returns 4 if the file exists and overwriting is not permitted

int writeImage(char* filename,AnalyzeImage* img,bool overwrite, bool signedData) 
{
  char imgfname[256];
 
  int imgsize;
  int i; 

 
  imgsize = (img->header.x_dim)*(img->header.y_dim)*(img->header.z_dim); 
  imagename(imgfname,filename);
  if(! overwrite) {
    ifstream ifile(imgfname,ios::binary);
    if(ifile.is_open()) {
      ifile.close();
      return(4);
    }
  }
  if(writeHeader(filename,img->header) == false) return(1);
  ofstream ofile(imgfname,ios::binary);
  //  else ofstream ofile(imgfname,ios::binary|ios::noreplace); 
  if (!ofile) return(2);
  switch(img->header.datatype) {
	case DT_UNSIGNED_CHAR:
	    for(i = 0;i<imgsize;i++) {
              unsigned char tmp;
	      img->data[i] = ROUND(img->data[i]);
              img->data[i] = MIN(img->data[i],255);
              img->data[i] = MAX(img->data[i],0);
              tmp = (unsigned char) img->data[i];
              ofile.write((char *)&tmp,sizeof(char)); 
            } 
            break;
  	case DT_SIGNED_SHORT:
	    for(i = 0;i<imgsize;i++) {
              if(signedData) {
                short tmp;
	        img->data[i] = ROUND(img->data[i]);
                img->data[i] = MIN(img->data[i],32768);
                img->data[i] = MAX(img->data[i],-32767);
                tmp = (short) img->data[i];
                ofile.write((char *)&tmp,sizeof(short));
              }
              else {
                unsigned short tmp;
                img->data[i] = ROUND(img->data[i]);
                img->data[i] = MIN(img->data[i],65535);
                img->data[i] = MAX(img->data[i],0);
                tmp = (unsigned short) img->data[i];
                ofile.write((char *)&tmp,sizeof(short));
	      }
            } 
            break;
        case DT_SIGNED_INT:
	    for(i = 0;i<imgsize;i++) {
              int tmp;
	      img->data[i] = ROUND(img->data[i]);
              img->data[i] = MIN(img->data[i],2147000000);
              img->data[i] = MAX(img->data[i],-2147000000);
              tmp = (int) img->data[i];
              ofile.write((char *)&tmp,sizeof(int)); 
            }   
            break;
        case DT_FLOAT:
	    for(i = 0;i<imgsize;i++) {
              float tmp;
              tmp =  img->data[i];
              ofile.write((char *)&tmp,sizeof(float)); 
            }
            break; 
  default: return(3); break;
  ofile.close();

  }   
  return(0);
}

int readLabelImage(char* filename,AnalyzeLabelImage* img)
{
  char imgfname[256];
  unsigned char* tmpchar;
  int i,imgsize;
  bool bs;
 
  if(readHeader(filename,img->header) == false) return(1);
  bs = byteSwapNecessary(&img->header);
  if(bs) swapHeader(img->header);  

  imgsize = (img->header.x_dim)*(img->header.y_dim)*(img->header.z_dim); 
  // img->data = (float *) malloc(imgsize*sizeof(float));  
  img->data = new unsigned char[imgsize];
  if(img->data == NULL) return(2);
  imagename(imgfname,filename);
  ifstream ifile(imgfname,ios::binary);
  if (!ifile) return(3);
  if(img->header.datatype != DT_UNSIGNED_CHAR) return(4);
  else {
    ifile.read((char*) img->data,imgsize);
    ifile.close();
    return(0); 
  }
}

int writeLabelImage(char* filename,AnalyzeLabelImage* img,bool overwrite)
{
  char imgfname[256];
 
  int imgsize;
  int i; 

  imgsize = (img->header.x_dim)*(img->header.y_dim)*(img->header.z_dim);
  imagename(imgfname,filename); 
  if(! overwrite) {
    ifstream ifile(imgfname,ios::binary);
    if(ifile.is_open()) {
      ifile.close();
      return(4);
    }
  }
  if(writeHeader(filename,img->header) == false) return(1);  
  ofstream ofile(imgfname,ios::binary);
  if (!ofile) return(2);
  ofile.write((char *)img->data,sizeof(char)*imgsize);
  ofile.close();
  return(0);

}

// bool copyHeader(AnalyzeHeader* source, AnalyzeHeader* target);

bool copyImage(AnalyzeImage* source,AnalyzeImage* target) 
{
  int imgsize,i;

  target->header = source->header;
  imgsize = (source->header.x_dim)*(source->header.y_dim)*(source->header.z_dim);
  target->data = new float [imgsize];
  if(target->data == NULL) return false;
  for(i = 0;i < imgsize;i++) {
    target->data[i] = source->data[i];
  }
  return true;
}

bool copyLabelImage(AnalyzeLabelImage* source,AnalyzeLabelImage* target)
{
  int imgsize,i;

  target->header = source->header;
  imgsize = (source->header.x_dim)*(source->header.y_dim)*(source->header.z_dim);
  target->data = new unsigned char [imgsize];
  if(target->data == NULL) return false;
  for(i = 0;i < imgsize;i++) {
    target->data[i] = source->data[i];
  }
  return true;
}

bool newImage(AnalyzeImage* img, AnalyzeImage* alike) 
{
  int imgsize,i;

  img->header = alike->header;
  imgsize = (alike->header.x_dim)*(alike->header.y_dim)*(alike->header.z_dim);
  img->data = new float [imgsize];
  if(img->data == NULL) return false;
  for(i = 0;i < imgsize;i++) {
    img->data[i] = 0;
  }
  return true;
}

bool newLabelImage(AnalyzeLabelImage* img, AnalyzeImage* alike)
{
  int imgsize,i;

  img->header = alike->header;
  img->header.datatype = DT_UNSIGNED_CHAR;
  imgsize = (alike->header.x_dim)*(alike->header.y_dim)*(alike->header.z_dim);
  img->data = new unsigned char [imgsize];
  if(img->data == NULL) return false;
  for(i = 0;i < imgsize;i++) {
    img->data[i] = 0;
  }
  return true;
}

bool mask2Image(AnalyzeLabelImage* source, AnalyzeImage* target) 
{
{
  int imgsize,i;

  target->header = source->header;
  imgsize = (source->header.x_dim)*(source->header.y_dim)*(source->header.z_dim);
  target->data = new float [imgsize];
  if(target->data == NULL) return false;
  for(i = 0;i < imgsize;i++) {
    target->data[i] = source->data[i];
  }
  return true;
}
}

bool binaryMask(AnalyzeImage* img,AnalyzeLabelImage* mask,float tr) 
{
  int i;
  int imgsize; 

  if(!newLabelImage(mask,img)) return(false);
  imgsize = (img->header.x_dim)*(img->header.y_dim)*(img->header.z_dim); 
  for(i = 0;i < imgsize;i++) {
    if(img->data[i] > tr) mask->data[i] = 1;
    else mask->data[i] = 0;
  }
  return(true);
}

// finds the closest voxel with the value greater than tr to the voxel (x,y,z)
// extremely slow but correct 
// returns false if no voxel value exceeds the threshold

bool findClosestNonZero(AnalyzeImage* img,int x,int y, int z,int* cx, int* cy, int* cz,float tr)
{
  int i,j,k;
  float d;
  bool found; 
  int windowSize = 11;
  // Initialize
  *cx = x;  // closest x coordinate
  *cy = y;  // closest y coordinate
  *cz = z;  // closest z coordinate
  found = false;
  // if value at (x,y,z) > tr then ok
  if(getVoxelValue(img,*cx,*cy,*cz) > tr) return(true);
  d = pow((float) (img->header.x_dim + img->header.y_dim  + img->header.z_dim),2);
  for(i = (-windowSize);i < (windowSize + 1);i++) {
    for(j = (-windowSize);j < (windowSize + 1) ;j++) {
      for(k = (-windowSize);k < (windowSize + 1);k++) {
         if(getSafeVoxelValue(img,x + i,y + j,z + k) > tr) {
           if( (pow((float) i,2) + pow((float) j,2) + pow((float) k,2)) < d) {
             d = (pow((float) i,2) + pow((float) j ,2) + pow((float) k,2));
             *cx = x + i;
             *cy = y + j;
             *cz = z + k;
             found = true;
        
           }
         }
       }
     }
   }
   return(found);

}
void averageFilterZeros(AnalyzeImage* img,AnalyzeImage* refImg,float tr,int times)
{
  int x,y,z,i;
  int x1,y1,z1;
  float tmpf;  

  for(i = 0;i < times;i++) {
    for(x = 0;x < img->header.x_dim;x++) {
      for(y = 0;y < img->header.y_dim;y++) {
        for(z = 0;z < img->header.z_dim;z++) {
          if(getVoxelValue(refImg,x,y,z) < tr) {
	    tmpf = 0.0;
            for(x1 = -3;x1 < 4;x1++) {
              for(y1 = -3;y1 < 4;y1++) {
                for(z1 = -3;z1 < 4;z1++) {
                  tmpf = tmpf + getSafeVoxelValue(img,x + x1,y + y1,z + z1);
		}
              }
            }
            putVoxelValue(img,x,y,z,tmpf/(7*7*7));
	  }
	}
      }
    }
  } 
}

// finds a value such that exatcly a certain percentage of voxel values (in img) 
// are smaller than it. Considers only voxels within the mask.

float imageKth(AnalyzeImage* img, AnalyzeLabelImage* mask, float percentage)
{
  long int i,j,k,l,m,n;
  int imgSize;
  float x;
  float* a; 
  
  imgSize = (img->header.x_dim)*(img->header.y_dim)*(img->header.z_dim);
  n = 0;
  for(i = 0;i < imgSize;i++) {
    if(mask->data[i] > 0.5) {
      n++;
    }
  }
  k = (long int) floor(percentage*n);
  if(k > n) return(0);
  if(k < 0) return(0);
  a = new float[n];
  n = 0;
  for(i = 0;i < imgSize;i++) {
    if(mask->data[i] > 0.5) {
      a[n] = img->data[i];
      n++;
    }
  }
  
 l=0 ; m=n-1 ;
 while (l<m) {
   x=a[k] ;
   i=l ;
   j=m ;
   do {
     while (a[i]<x) i++ ;
     while (x<a[j]) j-- ;
     if (i<=j) {
       FLOAT_SWAP(a[i],a[j]) ;
       i++ ; j-- ;
     }
   } while (i<=j) ;
   if (j<k) l=i ;
   if (k<i) m=j ;
 }
 x = a[k];
 delete[] a;
 return(x);
}

// Median filters image img; 
// Assumes windowLengths are odd integers
// for example (medianFilter3D(..., ..., 1, 1, 1);
// uses 3 x 3 x 3 mask.

void medianFilter3D(AnalyzeImage* img, AnalyzeLabelImage* mask, int windowLenX, int windowLenY, int windowLenZ)
{
  int windowSize;
  float* window;
  int x,y,z,i,j,k,l,m;
  int n;
  int med;
  float f;

  windowSize = (2*windowLenX + 1)*(2*windowLenY + 1)*(2*windowLenZ + 1);  
  med = (windowSize - 1)/2;
  window = new float[windowSize];


  for(x = 0;x < img->header.x_dim;x++) {
    for(y = 0;y < img->header.y_dim;y++) {
      for(z = 0;z < img->header.z_dim;z++) {
        if(getLabelValue(mask,x,y,z) > 0.5) {
          n = 0;
          for(i = (-windowLenX);i < (windowLenX + 1);i++) {
            for(j = (-windowLenY);j < (windowLenX + 1);j++) {
              for(k = (-windowLenZ);k < (windowLenX + 1);k++) {
                window[n] = getSafeVoxelValue(img,x + i, y + j, z + k);
                n++;
              }
            }
          }
          l=0 ; m=n-1 ;
          while (l<m) {
            f=window[med] ;
            i=l ;
            j=m ;
            do {
              while (window[i]<f) i++ ;
              while (f<window[j]) j-- ;
              if (i<=j) {
                FLOAT_SWAP(window[i],window[j]) ;
                i++ ; j-- ;
              }
            } while (i<=j) ;
            if (j<med) l=i ;
            if (med<i) m=j ;
	  }
	  putVoxelValue(img,x,y,z,window[med]);
        }
      }
    }
  }
  delete[] window;
}

void knnFilter(AnalyzeImage* img, AnalyzeLabelImage* mask, int windowLenX, int windowLenY, int windowLenZ, int knn)
{
  float* window;
  float* distances;
  int x,y,z,i,j,k;
  int n;
  float f,sum,voxelVal;
  char c;

  window = new float[knn]; 
  distances = new float[knn];

  for(x = 0;x < img->header.x_dim;x++) {
    for(y = 0;y < img->header.y_dim;y++) {
      for(z = 0;z < img->header.z_dim;z++) {
        if(getLabelValue(mask,x,y,z) > 0.5) {
          n = 0;
          voxelVal = getVoxelValue(img,x,y,z);
          for(i = (-windowLenX);i < (windowLenX + 1);i++) {
            for(j = (-windowLenY);j < (windowLenX + 1);j++) {
              for(k = (-windowLenZ);k < (windowLenX + 1);k++) { 
		if(n < knn) {
                  window[n] = getSafeVoxelValue(img,x + i, y + j, z + k);
                  distances[n] = fabs(window[n] - voxelVal);
                   n++;
                }
                else {
                  c = maxArg(distances,knn);
                  f = fabs(getSafeVoxelValue(img,x + i, y + j, z + k) - voxelVal);
                  if(f < distances[c]) {
                    distances[c] = f;
                    window[c] = getSafeVoxelValue(img,x + i, y + j, z + k);
		  }
                }
              }
            }
          }
          sum = 0.0;
          for(i = 0;i < knn;i++) {
            sum = sum + window[i];
          }
          putVoxelValue(img,x,y,z,sum/((float) knn));
        }
      }  
    }
  }
  delete[] window;
  delete[] distances;
}
